/// Simple optionally-singleton service locator
///
/// Use [GotIt.registerFactory], [GotIt.registerSingleton]
/// or [GotIt.registerLazySingleton] to register whatever you desire.
///
/// Use [GotIt.get] or call the object like a method ([GotIt.call])
/// to get whatever you registered back.
class GotIt {
  /// Creates an unique instance of [GotIt]
  GotIt();

  /// Returns a singleton instance of [GotIt]
  factory GotIt.singleton() => _instance;

  static final _instance = GotIt();

  // DI/SL Containers are always just Maps, this one isn't a singleton however.
  final _types = <_TypeWithString, _Entry>{};

  /// Register a method that will return a new instance of T
  /// everytime [get] (or [call]) is called.
  void registerFactory<T>(
    T Function() func, {
    bool allowReassignment = false,
    String instanceName,
  }) =>
      _register<T>(allowReassignment, false, func, instanceName);

  /// Register an item that will be returned verbatim
  /// everytime [get] (or [call]) is called.
  void registerSingleton<T>(
    T instance, {
    bool allowReassignment = false,
    String instanceName,
  }) =>
      _register<T>(allowReassignment, false, instance, instanceName);

  /// Register a method that will create an instance of T on the first call
  /// to [get] (or [call]), and return that instance from then on.
  void registerLazySingleton<T>(
    T Function() func, {
    bool allowReassignment = false,
    String instanceName,
  }) =>
      _register<T>(allowReassignment, true, func, instanceName);

  // The backbone of the library
  void _register<T>(
    bool allowReassignment,
    bool isLazy,
    dynamic value,
    String instanceName,
  ) {
    final identifier = _TypeWithString(T, instanceName ?? '');
    assert(
      _types[identifier]?.isReassignable ?? true,
      'You cannot reassign to this type',
    );
    _types[identifier] = _Entry(allowReassignment, isLazy, value);
  }

  /// Same as [get]
  T call<T>([String instanceName = '']) => get<T>(instanceName);

  /// Grab and return whatever is associated with T
  T get<T>([String instanceName = '']) {
    final identifier = _TypeWithString(T, instanceName);
    final entry = _types[identifier];

    if (entry == null) {
      throw UnregisteredError<T>();
    }

    if (entry.value is T) {
      return entry.value;
    }

    if (entry.isLazy) {
      _types[identifier] = entry.activate<T>();
      return _types[identifier].value;
    }

    return _types[identifier].value();
  }

  /// Delete all registrations, useful for unit tests.
  void reset() => _types.clear();
}

class _TypeWithString {
  const _TypeWithString(this.type, this.string);

  final Type type;
  final String string;

  @override
  // TODO: implement hashCode
  int get hashCode => type.hashCode + string.hashCode;

  @override
  bool operator ==(Object other) =>
      other is _TypeWithString && other.type == type && other.string == string;
}

// Contains all entries and represents their behavior at [GotIt.get] time.
class _Entry {
  const _Entry(this.isReassignable, this.isLazy, this.value);

  final bool isReassignable;
  final bool isLazy;
  // This is explicitly dynamic because Dart has no ADT,
  // otherwise, it would be "T Function() | T"
  final dynamic value;

  /// If the current entry is a lazy singleton, instantiate it.
  _Entry activate<T>() {
    if (value is T Function() && isLazy) {
      return _Entry(isReassignable, false, value());
    } else {
      throw StateError(
        'Attempted to activate a non-lazy or non-function value',
      );
    }
  }
}

/// Thrown whenever a call to [GotIt.get] (or [GotIt.call]) fails to resolve T.
class UnregisteredError<T> extends Error {
  @override
  String toString() => "Type $T hasn't been registered.";
}
